import 'dart:async';
import 'package:meta/meta.dart';
import 'package:better_player/better_player.dart';
import 'package:flutter/material.dart';
import 'tivio_player_loader.dart';
import 'controls/better_player_ad_controls.dart';
import 'controls/better_player_control.dart';
import 'controls/better_player_ui_press_controls.dart';
import '../tivio.dart';
import '../player/tivio_player_input_source.dart';
import '../player/tivio_player_wrapper.dart';
import '../player/tivio_player_controller.dart';
import '../logger.dart';
import '../enums/tivio_player_state.dart';

class TivioPlayer extends StatefulWidget {
  String playerId;
  Tivio tivio;
  TivioPlayerController controller;
  TivioDataVideo initialTivioVideoSource;
  bool startInFullscreen;

  TivioPlayer({
    @required String playerId,
    @required Tivio tivio,
    TivioPlayerController controller,
    TivioDataVideo source,
    bool startInFullscreen = false,
  }) {
    this.playerId = playerId;
    this.tivio = tivio;
    this.controller = (controller == null) ? TivioPlayerController() : controller;
    this.initialTivioVideoSource = source;
    this.startInFullscreen = startInFullscreen;
  }

  @override
  _TivioPlayerState createState() => _TivioPlayerState();
}

class _TivioPlayerState extends State<TivioPlayer> {
  final Logger logger = Logger();
  bool isBetterPlayerInitialized = false;
  TivioPlayerWrapper tivioPlayerWrapper;
  BetterPlayerController betterPlayerController;
  StreamController<BetterPlayerEvent> playerEventStream;
  int seekToAfterInitialization;
  Map<String, BetterPlayerDataSource> sources = Map();
  TivioPlayerState playerState = TivioPlayerState.idle;

  @override
  void initState() {
    super.initState();
    synchronousInitState();
  }

  @override
  Widget build(BuildContext context) {
    return Container(
      color: Colors.black,
      child: (isBetterPlayerInitialized)
          ? BetterPlayer(controller: betterPlayerController)
          : TivioPlayerLoader(fullscreen: widget.startInFullscreen),
    );
  }

  @override
  void dispose() {
    if (playerEventStream != null) {
      playerEventStream.close();
    }
    betterPlayerController = null;
    super.dispose();
  }

  /// Public

  openAndPlay(TivioDataVideo tivioDataVideo) {
    if (tivioDataVideo.isPlaylist) {
      throw new Exception('Playlists are currently unsupported');
    }

    tivioPlayerWrapper.openAndPlay(TivioPlayerVodSource(
      id: tivioDataVideo.id,
      uri: tivioDataVideo.url.hls,
      name: tivioDataVideo.name,
      description: '',
    ));
  }

  /// Player wrapper integration

  playerPlay(payload) {
    betterPlayerController.play();
  }

  playerPause(payload) {
    betterPlayerController.pause();
  }

  playerSeekTo(Duration time) {
    betterPlayerController.seekTo(time);
  }

  playerOpenAndPlay(TivioPlayerInputSource source, {int currentTime}) async {
    logger.verbose('Open and play source id: ${source.id}, type: ${source.type}');

    seekToAfterInitialization = currentTime;

    bool isSourceAd = (source.type == TivioPlayerWrapperInputSourceType.Advertisement);
    setupBetterPlayer(source, isSourceAd);
  }

  /// Private

  skipAd() {
    tivioPlayerWrapper.skipAd();
  }

  synchronousInitState() async {
    await initTivioPlayerWrapper();
    initTivioPlayerWrapperController();

    if (this.widget.initialTivioVideoSource != null) {
      openAndPlay(this.widget.initialTivioVideoSource);
    }
  }

  initTivioPlayerWrapper() async {
    tivioPlayerWrapper = await widget.tivio.getPlayer(widget.playerId);

    tivioPlayerWrapper.register(TivioPlayerWrapperController(
      playerPlay,
      playerPause,
      playerSeekTo,
      playerOpenAndPlay,
    ));
  }

  initTivioPlayerWrapperController() {
    widget.controller.play = tivioPlayerWrapper.play;
    widget.controller.pause = tivioPlayerWrapper.pause;
    widget.controller.openAndPlay = openAndPlay;
    widget.controller.seekTo = tivioPlayerWrapper.seekTo;
  }

  setupBetterPlayer(TivioPlayerInputSource tivioSource, bool isSourceAd) {
    if (betterPlayerController != null) {
      betterPlayerController.pause();
      betterPlayerController.removeEventsListener(this.handleBetterPlayerEvent);
    }

    if (playerEventStream != null) {
      playerEventStream.close();
    }

    var newPlayerEventStream = StreamController<BetterPlayerEvent>.broadcast();
    var config = BetterPlayerConfiguration(
      aspectRatio: 16 / 9,
      fit: BoxFit.fill,
      autoPlay: false,
      autoDispose: true,
      fullScreenByDefault: true, //widget.startInFullscreen,
      controlsConfiguration: (isSourceAd)
          ? BetterPlayerControlsConfiguration(
              playerTheme: BetterPlayerTheme.custom,
              customControlsBuilder: (controller) => BetterPlayerAdControls(
                eventPlayerStream: newPlayerEventStream,
                source: tivioSource,
                onClick: () => skipAd(),
              ),
            )
          : BetterPlayerControlsConfiguration(
              playerTheme: BetterPlayerTheme.custom,
              customControlsBuilder: (controller) => BetterPlayerControl(
                controller: controller,
                eventPlayerStream: newPlayerEventStream,
                source: tivioSource,
                controls: BetterPlayerUiPressControls(
                  onPlayPressed: () => widget.controller.play(),
                  onPausePressed: () => widget.controller.pause(),
                  onForwardPressed: () async {
                    Duration position = await controller.videoPlayerController.position;
                    widget.controller.seekTo(Duration(seconds: position.inSeconds + 15));
                  },
                  onBackwardPressed: () async {
                    Duration position = await controller.videoPlayerController.position;
                    widget.controller.seekTo(Duration(seconds: position.inSeconds - 15));
                  },
                  onEnterFullscreenPressed: () => controller.enterFullScreen(),
                  onExitFullscreenPressed: () => controller.exitFullScreen(),
                  onSeekingStarted: () => widget.controller.pause(),
                  onSeekingEnded: (value) {
                    Duration sourceDuration = controller.videoPlayerController.value.duration;
                    int seekToSecond = (sourceDuration.inSeconds * value).ceil();
                    widget.controller.seekTo(Duration(seconds: seekToSecond));
                  },
                ),
              ),
            ),
    );

    var source = createPlayerSource(tivioSource);
    var controller = BetterPlayerController(config);

    setState(() {
      betterPlayerController = controller;
      betterPlayerController.setupDataSource(source);
      betterPlayerController.addEventsListener(this.handleBetterPlayerEvent);
      betterPlayerController.pause();

      isBetterPlayerInitialized = true;
      playerEventStream = newPlayerEventStream;
    });
  }

  BetterPlayerDataSource createPlayerSource(TivioPlayerInputSource tivioPlayerInputSource) {
    switch (tivioPlayerInputSource.type) {
      case TivioPlayerWrapperInputSourceType.Vod:
        if (sources[tivioPlayerInputSource.id] != null) {
          return sources[tivioPlayerInputSource.id];
        }

        var source = BetterPlayerDataSource(
          BetterPlayerDataSourceType.network,
          tivioPlayerInputSource.uri,
          cacheConfiguration: BetterPlayerCacheConfiguration(
            useCache: true,
            preCacheSize: 100 * 1024 * 1024,
            maxCacheSize: 100 * 1024 * 1024,
            maxCacheFileSize: 100 * 1024 * 1024,
          ),
        );

        sources[tivioPlayerInputSource.id] = source;
        return source;
      case TivioPlayerWrapperInputSourceType.Advertisement:
        return BetterPlayerDataSource(
          BetterPlayerDataSourceType.network,
          tivioPlayerInputSource.uri,
        );
      default:
        throw new Exception('Unsupported type ${tivioPlayerInputSource.type} in createPlayerSource');
    }
  }

  handleBetterPlayerEvent(BetterPlayerEvent event) {
    if (playerEventStream != null) {
      playerEventStream.add(event);
    }

    switch (event.betterPlayerEventType) {
      case BetterPlayerEventType.initialized:
        if (seekToAfterInitialization != null) {
          playerSeekTo(Duration(milliseconds: seekToAfterInitialization));
        }
        if (betterPlayerController != null) {
          betterPlayerController.play();
        } else {
          logger.error('Unable to play video. Player is not initialized.');
        }
        break;
      case BetterPlayerEventType.play:
        tivioPlayerWrapper.onStateChanged(TivioPlayerState.playing);
        break;
      case BetterPlayerEventType.pause:
        tivioPlayerWrapper.onStateChanged(TivioPlayerState.paused);
        break;
      case BetterPlayerEventType.progress:
        tivioPlayerWrapper.onTimeChanged(event.parameters['progress']);
        break;
      case BetterPlayerEventType.seekTo:
        tivioPlayerWrapper.onTimeChanged(event.parameters['duration']);
        break;
      case BetterPlayerEventType.finished:
        tivioPlayerWrapper.onPlaybackEnded();
        // if current source = vod -> call onEnded
        // if current souurce = ad & is last -> call onEnded
        break;
      default:
        break;
    }
  }
}
